// 
// Copyright (c) 2009, Daisuke Deguchi, Masafumi Noda
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the Nagoya University nor the names of its contributors
// may be used to endorse or promote products derived from this software
// without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
// IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
// THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 

#include <iostream>
#include <cmath>

#include <mist/mist.h>
#include <mist/converter.h>
#include <mist/drawing.h>
#include <mist/timer.h>
#include <mist/io/bmp.h>

// RX^g̐錾
__constant__ int   __R__[ 6 ];			// KEVAtB^̃tB^a
__constant__ float __cf__[ 6 * 23 ];	// KEVAtB^̃tB^W
__constant__ int4  __mipmap__[ 20 ];		// \z~bv}bv̍WƃTCY
__constant__ int   __ori_size__[ 2 ];	// ͉摜̃TCY


// eNX`̐錾
texture< float, 1 > oriTex;				// ͉摜̃eNX`
texture< float, 1 > workTex;			// ƗpeNX`
texture< float2, 2 > imgTex;			// DoGvZ̉摜\eNX`


// KEVAtB^̔aƌW߂֐
inline void compute_gaussian_kernel( float cf[], int &R, double sigma )
{
	double radius = sigma * 2.0;

	R = static_cast< int >( std::ceil( radius ) );

	for( int i = 0 ; i < 23 ; i++ )
	{
		cf[ i ] = 0.0f;
	}

	int fw = R * 2 + 1;
	int rw = fw / 2;

	double _1_sig2 = 1.0 / ( sigma * sigma * 2.0 );
	double sum = 0.0, ccf[ 23 ];
	for( int x = -rw ; x <= rw ; x++ )
	{
		double xx = static_cast< double >( x * x );
		double g = std::exp( -xx * _1_sig2 );
		sum += g;
		ccf[ x + rw ] = g;
	}

	for( int i = 0 ; i < fw ; i++ )
	{
		cf[ i ] = static_cast< float >( ccf[ i ] / sum );
	}
}

// Xɑ΂ăKEVAtB^{֐
__global__ void kernel1( float *out, int w, int h, int W, int H, int scale, int oX, int oY )
{
	int si = threadIdx.x + blockDim.x * blockIdx.x;
	int sj = threadIdx.y + blockDim.y * blockIdx.y;

	// ͈͓̉fɑ΂Ă̂ݏs
	if( si < w && sj < h )
	{
		for( int k = 0 ; k < 6 ; k++ )
		{
			float *cf = &__cf__[ k * 23 ];
			int R = __R__[ k ];

			int s = si - R < 0  ? 0 - si : -R;
			int e = si + R >= w ? w - si :  R;

			// Xɑ΂ăKEVAtB^{
			float pix = 0.0f, sum = 0.0f;
			for( int n = s ; n < e ; n++ )
			{
				float g = cf[ n + R ];

				pix += tex1Dfetch( oriTex, ( ( si + n ) + sj * __ori_size__[ 0 ] ) * scale ) * g;
				sum += g;
			}

			out[ si + oX + ( sj + oY + k * H ) * W ] = pix / sum / 255.0f;
		}
	}
}

// Yɑ΂ăKEVAtB^{֐
__global__ void kernel2( float2 *out, int w, int h, int W, int H, int Pitch, int oX, int oY )
{
	int si = threadIdx.x + blockDim.x * blockIdx.x;
	int sj = threadIdx.y + blockDim.y * blockIdx.y;

	// ͈͓̉fɑ΂Ă̂ݏs
	if( si < w && sj < h )
	{
		int oi = si + oX;
		int oj = sj + oY;

		float o[ 6 ];
		for( int k = 0 ; k < 6 ; k++ )
		{
			float *cf = &__cf__[ k * 23 ];
			int R = __R__[ k ];

			int s = sj - R < 0  ? 0 - sj : -R;
			int e = sj + R >= h ? h - sj :  R;

			// Yɑ΂ăKEVAtB^{
			float pix = 0.0f, sum = 0.0f;
			for( int n = s ; n < e ; n++ )
			{
				float g = cf[ n + R ];

				pix += tex1Dfetch( workTex, oi + ( oj + n + k * H ) * W ) * g;
				sum += g; 
			}

			o[ k ] = pix / sum;
		}

		// KEVAtB^̌ʂDoGo͂
		( ( float2 * )( ( char * )out + ( oj + 0 * H ) * Pitch ) )[ oi ] = make_float2( o[ 0 ], o[ 1 ] - o[ 0 ] );
		( ( float2 * )( ( char * )out + ( oj + 1 * H ) * Pitch ) )[ oi ] = make_float2( o[ 1 ], o[ 2 ] - o[ 1 ] );
		( ( float2 * )( ( char * )out + ( oj + 2 * H ) * Pitch ) )[ oi ] = make_float2( o[ 2 ], o[ 3 ] - o[ 2 ] );
		( ( float2 * )( ( char * )out + ( oj + 3 * H ) * Pitch ) )[ oi ] = make_float2( o[ 3 ], o[ 4 ] - o[ 3 ] );
		( ( float2 * )( ( char * )out + ( oj + 4 * H ) * Pitch ) )[ oi ] = make_float2( o[ 4 ], o[ 5 ] - o[ 4 ] );
		( ( float2 * )( ( char * )out + ( oj + 5 * H ) * Pitch ) )[ oi ] = make_float2( o[ 5 ], 0.0f );
	}
}

// 召֌W𔻒肷֐iGPUł̂ݎsj
__device__ inline void minmax( float &min_, float &max_, float2 val )
{
	min_ = fminf( min_, val.y );
	max_ = fmaxf( max_, val.y );
}

// L[|Cgo֐
__global__ void kernel3( float4 *out, int w, int h, int W, int H, int oX, int oY, int sci, int scale )
{
	int si = threadIdx.x + blockDim.x * blockIdx.x;
	int sj = threadIdx.y + blockDim.y * blockIdx.y;

	// ͈͓̉fɑ΂Ă̂ݏs
	if( 0 < si && si < w - 1 && 0 < sj && sj < h - 1 && out[ ( si + sj * W ) * scale ].x == 0 )
	{
		int i = si + oX;
		int j = sj + oY;

		for( int k = 3 ; k > 0 ; k-- )
		{
			float p1 = tex2D( imgTex, i - 1, j - 1 + ( k - 1 ) * H ).y, p2 = p1;
			int k0   = ( k - 1 ) * H;
			int k1   = k * H;
			int k2   = ( k + 1 ) * H;

			minmax( p1, p2, tex2D( imgTex, i    , j - 1 + k0 ) );
			minmax( p1, p2, tex2D( imgTex, i + 1, j - 1 + k0 ) );
			minmax( p1, p2, tex2D( imgTex, i - 1, j     + k0 ) );
			minmax( p1, p2, tex2D( imgTex, i    , j     + k0 ) );
			minmax( p1, p2, tex2D( imgTex, i + 1, j     + k0 ) );
			minmax( p1, p2, tex2D( imgTex, i - 1, j + 1 + k0 ) );
			minmax( p1, p2, tex2D( imgTex, i    , j + 1 + k0 ) );
			minmax( p1, p2, tex2D( imgTex, i + 1, j + 1 + k0 ) );

			minmax( p1, p2, tex2D( imgTex, i - 1, j - 1 + k1 ) );
			minmax( p1, p2, tex2D( imgTex, i    , j - 1 + k1 ) );
			minmax( p1, p2, tex2D( imgTex, i + 1, j - 1 + k1 ) );
			minmax( p1, p2, tex2D( imgTex, i - 1, j     + k1 ) );
			minmax( p1, p2, tex2D( imgTex, i + 1, j     + k1 ) );
			minmax( p1, p2, tex2D( imgTex, i - 1, j + 1 + k1 ) );
			minmax( p1, p2, tex2D( imgTex, i    , j + 1 + k1 ) );
			minmax( p1, p2, tex2D( imgTex, i + 1, j + 1 + k1 ) );

			minmax( p1, p2, tex2D( imgTex, i - 1, j - 1 + k2 ) );
			minmax( p1, p2, tex2D( imgTex, i    , j - 1 + k2 ) );
			minmax( p1, p2, tex2D( imgTex, i + 1, j - 1 + k2 ) );
			minmax( p1, p2, tex2D( imgTex, i - 1, j     + k2 ) );
			minmax( p1, p2, tex2D( imgTex, i    , j     + k2 ) );
			minmax( p1, p2, tex2D( imgTex, i + 1, j     + k2 ) );
			minmax( p1, p2, tex2D( imgTex, i - 1, j + 1 + k2 ) );
			minmax( p1, p2, tex2D( imgTex, i    , j + 1 + k2 ) );
			minmax( p1, p2, tex2D( imgTex, i + 1, j + 1 + k2 ) );

			// ɒl̂L[|CǧƂ
			float pix0  = tex2D( imgTex, i, j + k1 ).y;
			if( pix0 < p1 || p2 < pix0 )
			{
				float pix1  = tex2D( imgTex, i + 1, j     + k1 ).y;
				float pix2  = tex2D( imgTex, i + 1, j - 1 + k1 ).y;
				float pix3  = tex2D( imgTex, i    , j - 1 + k1 ).y;
				float pix4  = tex2D( imgTex, i - 1, j - 1 + k1 ).y;
				float pix5  = tex2D( imgTex, i - 1, j     + k1 ).y;
				float pix6  = tex2D( imgTex, i - 1, j + 1 + k1 ).y;
				float pix7  = tex2D( imgTex, i    , j + 1 + k1 ).y;
				float pix8  = tex2D( imgTex, i + 1, j + 1 + k1 ).y;

				float Dxx   = pix1 + pix5 - 2.0f * pix0;
				float Dyy   = pix7 + pix3 - 2.0f * pix0;
				float Dxy   = ( pix8 + pix4 - pix2 - pix6 ) * 0.25;
				float Tr    = Dxx + Dyy;
				float Det   = Dxx * Dyy - Dxy * Dxy;
				float gamma = 10.0f;

				// GbW̓_O
				if( Det > 0 && Tr * Tr * gamma < Det * ( gamma + 1.0f ) * ( gamma + 1.0f ) )
				{
					float pix9  = tex2D( imgTex, i    , j     + k0 ).y;
					float pix10 = tex2D( imgTex, i    , j     + k2 ).y;
					float pix11 = tex2D( imgTex, i + 1, j     + k2 ).y;
					float pix12 = tex2D( imgTex, i - 1, j     + k0 ).y;
					float pix13 = tex2D( imgTex, i - 1, j     + k2 ).y;
					float pix14 = tex2D( imgTex, i + 1, j     + k0 ).y;
					float pix15 = tex2D( imgTex, i    , j + 1 + k2 ).y;
					float pix16 = tex2D( imgTex, i    , j - 1 + k0 ).y;
					float pix17 = tex2D( imgTex, i    , j - 1 + k2 ).y;
					float pix18 = tex2D( imgTex, i    , j + 1 + k0 ).y;

					float a11  = pix1 + pix5 - 2.0f * pix0;					// Dxx
					float a22  = pix3 + pix7 - 2.0f * pix0;					// Dyy
					float a33  = pix10 + pix9 - 2.0f * pix0;				// Dss
					float a12  = Dxy;										// Dxy
					float a13  = ( pix11 + pix12 - pix13 - pix14 ) * 0.25f;	// Dxs
					float a23  = ( pix15 + pix16 - pix17 - pix18 ) * 0.25f;	// Dys

					float A11  = a22 * a33 - a23 * a23;
					float A12  = a23 * a13 - a12 * a33;
					float A13  = a12 * a23 - a22 * a13;
					float detA = a11 * A11 + a13 * A13 + a12 * A12;

					// ts񂪑݂ꍇ̂tsvZ
					if( detA != 0 )
					{
						// evf̒lvZ
						float A22 = a11 * a33 - a13 * a13;
						float A23 = a13 * a12 - a11 * a23;
						float A33 = a11 * a22 - a12 * a12;

						float dx  = ( pix1 - pix5 ) * 0.5f;
						float dy  = ( pix7 - pix3 ) * 0.5f;
						float ds  = ( pix10 - pix9 ) * 0.5f;

						float _1_detA = 1.0f / detA;
						float xx  = -( A11 * dx + A12 * dy + A13 * ds ) * _1_detA;
						float yy  = -( A11 * dx + A22 * dy + A23 * ds ) * _1_detA;
						float zz  = -( A13 * dx + A23 * dy + A33 * ds ) * _1_detA;

						// L[|CgʒũTusNZ{RgXg̒Ⴂ_̍폜
						if( pix0 + 0.5f * ( dx * xx + dy * yy + ds * zz ) > 0.03f )
						{
							out[ ( si + sj * W ) * scale ] = make_float4( si + xx, sj + yy, k - 1, sci );
						}
					}
				}
			}
		}
	}
}

// L[|Cg̃IGe[VZo֐
__global__ void kernel4( float *hist, float4 *keys, int nkeys, int H )
{
	int indx = threadIdx.x + blockDim.x * blockIdx.x;

	// L[|Cg̐s
	if( indx < nkeys )
	{
		int s  = ( int )keys[ indx ].w - 1;			// L[|Cg̃XP[
		int k  = ( int )keys[ indx ].z;				// L[|Cg
		int oX = __mipmap__[ s ].x;					// XP[ɑΉ~bv}bvXW
		int oY = __mipmap__[ s ].y;					// XP[ɑΉ~bv}bvYW
		int w  = __mipmap__[ s ].z;					// XP[ɑΉ~bv}bv̕
		int h  = __mipmap__[ s ].w;					// XP[ɑΉ~bv}bv̍

		// L[|Cg̒SWєaIGe[VZoɗp̈肷
		int i  = ( int )( keys[ indx ].x + 0.5 );
		int j  = ( int )( keys[ indx ].y + 0.5 );
		int R  = __R__[ k ];
		int sx = i - R < 0  ? 0 - i : -R;
		int ex = i + R >= w ? w - i :  R;
		int sy = j - R < 0  ? 0 - j : -R;
		int ey = j + R >= h ? h - j :  R;

		// IWĩXP[ł̍WZo
		i += oX;
		j += oY + k * H;

		// IGe[V߂z̏
		float *cf = &__cf__[ k * 23 ];
		float Hist[ 36 ];
		for( int l = 0 ; l < 36 ; l++ )
		{
			Hist[ l ] = 0.0f;
		}

		// L[|Cgӂ̋PxlpăIGe[VZo
		for( int y = sy ; y < ey ; y++ )
		{
			float gy = cf[ y + R ];
			for( int x = sx ; x < ex ; x++ )
			{
				float gx    = cf[ x + R ];
				float g     = gx * gy;
				float fu    = tex2D( imgTex, i + x + 1, j + y ).x - tex2D( imgTex, i + x - 1, j + y ).x;
				float fv    = tex2D( imgTex, i + x, j + y + 1 ).x - tex2D( imgTex, i + x, j + y - 1 ).x;
				float mag   = sqrtf( fu * fu + fv * fv );
				int   theta = ( int )( 18 + 36 + 0.5 + atan2f( fv, fu ) / 3.1415926535897932384626433832795f * 18.0f ) % 36;

				// L[|Cg̒S̋ɉďdݕt
				Hist[ theta ] += g * mag;
			}
		}

		// IGe[ṼqXgO𕽊
		float tmp[ 3 ];
		tmp[ 0 ] = Hist[ 35 ];
		for( int l = 0 ; l < 36 ; l++ )
		{
			tmp[ 1 ]  = Hist[ l ];
			tmp[ 2 ]  = Hist[ ( l + 1 ) % 36 ];
			Hist[ l ] = tmp[ 0 ] * 0.25f + tmp[ 1 ] * 0.5f + tmp[ 2 ] * 0.25f;
			tmp[ 0 ]  = tmp[ 1 ];
			tmp[ 1 ]  = tmp[ 2 ];
		}

		// IGe[V̐K
		float mH = Hist[ 0 ];
		for( int l = 1 ; l < 36 ; l++ )
		{
			mH = fmaxf( mH, Hist[ l ] );
		}

		// ʂo͂
		for( int l = 0 ; l < 36 ; l++ )
		{
			hist[ indx * 36 + l ] = Hist[ l ] / mH;
		}
	}
}

int main( int argc, char *argv[] )
{
	// ̓t@Cw肳Ă邩ǂ`FbN
	if( argc != 3 )
	{
		printf( "This program requires path of a input and a output images.\n" );
		return( 1 );
	}

	typedef mist::rgb< unsigned char > color_type;
	typedef mist::array2< color_type > color_image_type;
	typedef mist::array2< float >      gray_image_type;

	// ͉摜IuWFNgp̕ϐ
	color_image_type ori;
	gray_image_type img;

	// 摜ǂݍ
	if( !mist::read_bmp( ori, argv[ 1 ] ) )
	{
		printf( "Failed to read image from [%s].\n", argv[ 1 ] );
		return( 1 );
	}

	// J[摜O[XP[摜ɕϊ
	mist::convert( ori, img );

	{
		// KEVAtB^̐ݒ
		int   Rs[ 6 ];			// KEVAtB^̔a
		float cfs[ 6 * 23 ];	// KEVAtB^̌W

		// KEVAJ[lvZ
		for( size_t i = 0 ; i < 6 ; i++ )
		{
			double sigma = 1.6 * std::pow( 2, i / 3.0 );
			compute_gaussian_kernel( &cfs[ i * 23 ], Rs[ i ], sigma );
		}

		// KEVAtB^̌W]
		cudaMemcpyToSymbol( __R__, Rs, 6 * sizeof( int ) );
		cudaMemcpyToSymbol( __cf__, cfs, 6 * 23 * sizeof( float ) );
	}

	// ~bv}bv\zۂ̃p[^LϐCщ摜TCY
	int4   mipmap[ 20 ];
	int    nMaps = 0;
	int    imgW  = static_cast< int >( img.width( ) * 3 / 2 ) + ( ( img.width( )  % 2 ) == 0 ? 0 : 1 );
	int    imgH  = static_cast< int >( img.height( ) );
	int    workW = imgW * imgH * 6;
	size_t wPitch = 0;

	// GPUɓ͉摜p̃̈mۂCeNX`Ƀ}bsO
	float *pOri;
	int osize[ 2 ] = { static_cast< int >( img.width( ) ), static_cast< int >( img.height( ) ) };
	cudaMalloc( ( void ** )&pOri, img.size( ) * sizeof( float ) );
	cudaMemcpy( pOri, &img[ 0 ], img.size( ) * sizeof( float ), cudaMemcpyHostToDevice );
	cudaBindTexture( 0, oriTex, pOri );
	cudaMemcpyToSymbol( __ori_size__, osize, 2 * sizeof( int ) );

	// GPȔo͉摜IuWFNgp̕ϐ
	mist::array2< float4 > out( img.width( ), img.height( ) );
	float  *pH = new float[ 36 * 10000 ];
	float  *pHist  = NULL;
	float  *pWork1 = NULL;
	float2 *pWork2 = NULL;
	float4 *pOut   = NULL;
	cudaMalloc( ( void ** )&pOut,   out.size( ) * sizeof( float4 ) );
	cudaMalloc( ( void ** )&pWork1, workW * sizeof( float ) );
	cudaMallocPitch( ( void ** )&pWork2, &wPitch, imgW * sizeof( float2 ), imgH * 6 );
	cudaMalloc( ( void ** )&pHist,  32 * 10000 * sizeof( float ) );

	// o͉摜̏ƍƗpeNX`̃}bsO
	cudaMemset( pOut, 0, out.size( ) * sizeof( float4 ) );
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc< float2 >( );
	cudaBindTexture( 0, workTex, pWork1 );
	cudaBindTexture2D( 0, imgTex, pWork2, channelDesc, imgW, imgH * 6, wPitch );


	// vZԂ̌vJn
	mist::timer t;

	{
		// KEVAs~bh\z

		int scale  = 1;												// KEVAs~bh\zɂXP[\ϐ
		int oX  = 0, oY = 0;										// KEVAtB^̏o͐W
		int width  = static_cast< int >( img.width( ) ) / scale;	// ݂̃XP[ɂ摜̕
		int height = static_cast< int >( img.height( ) ) / scale;	// ݂̃XP[ɂ摜̍

		for( size_t k = 0 ; ; k++ )
		{
			// sGPŨXbhw
			int threadNumX = 16;
			int threadNumY = 16;
			int blockNumX  = width  / threadNumX + ( ( width  % threadNumX ) == 0 ? 0 : 1 );
			int blockNumY  = height / threadNumY + ( ( height % threadNumY ) == 0 ? 0 : 1 );
			dim3 nThreads( threadNumX, threadNumY, 1 );
			dim3 nBlocks( blockNumX, blockNumY, 1 );

			// XɃKEVAtB^Kp
			kernel1<<< nBlocks, nThreads >>>( pWork1, width, height, imgW, imgH, scale, oX, oY );

			// ܂őҋ@
			cudaThreadSynchronize( );

			// YɃKEVAtB^Kp
			kernel2<<< nBlocks, nThreads >>>( pWork2, width, height, imgW, imgH, wPitch, oX, oY );

			// ܂őҋ@
			cudaThreadSynchronize( );

			// ݂̃XP[񂨂э\z~bv}bṽp[^L
			mipmap[ nMaps++ ] = make_int4( oX, oY, width, height );

			if( k == 0 )
			{
				oX += width;
			}
			else
			{
				oY += height;
			}

			width  /= 2;
			height /= 2;
			scale  *= 2;

			if( width < 23 || height < 23 )
			{
				break;
			}
		}
	}

	{
		// L[|Cgos
		for( int k = 0, s = 1 ; k < nMaps ; k++, s *= 2 )
		{
			// ~bv}bv̏擾
			int oX     = mipmap[ k ].x;
			int oY     = mipmap[ k ].y;
			int width  = mipmap[ k ].z;
			int height = mipmap[ k ].w;

			// sGPŨXbhw
			int threadNumX = 16;
			int threadNumY = 16;
			int blockNumX  = width  / threadNumX + ( ( width  % threadNumX ) == 0 ? 0 : 1 );
			int blockNumY  = height / threadNumY + ( ( height % threadNumY ) == 0 ? 0 : 1 );
			dim3 nThreads( threadNumX, threadNumY, 1 );
			dim3 nBlocks( blockNumX, blockNumY, 1 );

			// ݂̃XP[ŃL[|Cgos
			kernel3<<< nBlocks, nThreads >>>( pOut, width, height, out.width( ), imgH, oX, oY, k + 1, s );

			// ܂őҋ@
			cudaThreadSynchronize( );
		}
	}

	// GPȔʂCPUɓ]
	cudaMemcpy( &out[ 0 ], pOut, out.size( ) * sizeof( float4 ), cudaMemcpyDeviceToHost );

	// \z~bv}bv̏]
	cudaMemcpyToSymbol( __mipmap__, mipmap, nMaps * sizeof( int4 ) );

	// L[|Cg̃Xg쐬
	size_t nkeys = 0;
	for( size_t i = 0 ; i < out.size( ) ; i++ )
	{
		if( out[ i ].w > 0 )
		{
			out[ nkeys++ ] = out[ i ];
		}
	}

	// L[|Cg̃XgGPUɓ]
	cudaMemcpy( pOut, &out[ 0 ], nkeys * sizeof( float4 ), cudaMemcpyHostToDevice );

	// L[|Cg̃IGe[VZo
	{
		// sGPŨXbhw
		int threadNumX = 4 * 4;
		int blockNumX  = nkeys  / threadNumX + ( ( nkeys  % threadNumX ) == 0 ? 0 : 1 );
		dim3 nThreads( threadNumX, 1, 1 );
		dim3 nBlocks( blockNumX, 1, 1 );

		// GPUX̕s
		kernel4<<< nBlocks, nThreads >>>( pHist, pOut, nkeys, imgH );

		// GPȔʂ]
		cudaMemcpy( pH, pHist, nkeys * 36 * sizeof( float ), cudaMemcpyDeviceToHost );
	}

	printf( "Computation time: %.3f sec\n", t.elapse( ) );

	// oL[|Cg̈ʒuƃIGe[V`悷
	color_image_type tmp( ori );
	for( size_t i = 0 ; i < nkeys ; i++ )
	{
		int s = static_cast< int >( std::pow( 2.0, out[ i ].w - 1.0 ) + 0.5 );
		int x = static_cast< int >( out[ i ].x * s + 0.5 );
		int y = static_cast< int >( out[ i ].y * s + 0.5 );
		int r = static_cast< int >( 1.6 * std::pow( 2, ( out[ i ].z + 1.0 ) / 3.0 ) * s + 0.5 );
		mist::draw_circle( tmp, x, y, r, mist::colors< color_type >::yellow() );

		for( size_t j = 0 ; j < 36 ; j++ )
		{
			if( pH[ i * 36 + j ] >= 0.99f )
			{
				double theta = ( j - 18 ) / 18.0 * 3.1415926535897932384626433832795;
				mist::draw_line( tmp, x, y, ( int )( x + r * std::cos( theta ) ), ( int )( y + r * std::sin( theta ) ), mist::colors< color_type >::yellow() );
			}
		}
	}

	// ʂ̉摜ۑ
	if( !mist::write_bmp( tmp, argv[ 2 ] ) )
	{
		printf( "Failed to save image to [%s].\n", argv[ 2 ] );
		return( 1 );
	}

	// ̌n
	cudaUnbindTexture( oriTex );
	cudaUnbindTexture( imgTex );
	cudaUnbindTexture( workTex );
	cudaFree( pOri );
	cudaFree( pWork1 );
	cudaFree( pWork2 );
	cudaFree( pOut );
	cudaFree( pHist );
	delete [] pH;

	return( 0 );
}
