// 
// Copyright (c) 2009, Daisuke Deguchi, Masafumi Noda
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. Neither the name of the Nagoya University nor the names of its contributors
// may be used to endorse or promote products derived from this software
// without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
// IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
// THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 

#include <iostream>
#include <mist/mist.h>
#include <mist/drawing.h>
#include <mist/timer.h>
#include <mist/io/bmp.h>

#include "vector.h"


#define SCALE_FACTOR 1.2f   // ͉摜̊g嗦ݒ
#define MIN_SCALE 0.3f      // ŏ̊g嗦ݒ
#define MAX_SCALE 1.8f      // ő̊g嗦ݒ

// eNX`̐錾
texture< uchar4, 2, cudaReadModeElementType > imgTex;
texture< uchar4, 2, cudaReadModeNormalizedFloat > refTex;

// eNX`摜pz
cudaArray *imgArray;
cudaArray *refArray;


__global__ void kernel( float *error, float *scale, int imgW, int imgH, int areaW, int areaH, int maskW, int maskH, float s )
{
	int i = threadIdx.x + blockDim.x * blockIdx.x;
	int j = threadIdx.y + blockDim.y * blockIdx.y;

	float err = 0.0f;

	// ͈͓̉fɑ΂Ă̂ݏs
	if( i < areaW && j < areaH )
	{
		// sĂ݂eXg
		float _1_w = 1.0f / maskW;
		float _1_h = 1.0f / maskH;
		for( int n = 0 ; n < maskH ; n++ )
		{
			for( int m = 0 ; m < maskW ; m++ )
			{
				uchar4 p1 = tex2D( imgTex, i + m, j + n );
				float4 p2 = tex2D( refTex, _1_w * m, _1_h * n ) * 255.0f;
				err += ( p1.x - p2.x ) * ( p1.x - p2.x );
				err += ( p1.y - p2.y ) * ( p1.y - p2.y );
				err += ( p1.z - p2.z ) * ( p1.z - p2.z );
			}
		}

		err *= _1_w * _1_h;

		if( error[ i + j * imgW ] > err )
		{
			error[ i + j * imgW ] = err;
			scale[ i + j * imgW ] = s;
		}
	}
}


int main( int argc, char *argv[] )
{
	// ̓t@Cw肳Ă邩ǂ`FbN
	if( argc != 4 )
	{
		printf( "This program requires three parameters (input image, reference image, output image).\n" );
		return( 1 );
	}

	// ͉摜IuWFNgp̕ϐ
	typedef mist::rgba< unsigned char > pixel_type;
	mist::array2< pixel_type > img, ref;

	// 摜ǂݍ
	if( !mist::read_bmp( img, argv[ 1 ] ) )
	{
		printf( "Failed to read input image from [%s].\n", argv[ 1 ] );
		return( 1 );
	}

	// 摜ǂݍ
	if( !mist::read_bmp( ref, argv[ 2 ] ) )
	{
		printf( "Failed to read reference image from [%s].\n", argv[ 2 ] );
		return( 1 );
	}

	// GPUɓ͉摜p̃̈mۂCeNX`Ƀ}bsO
	cudaChannelFormatDesc channelDesc1 = cudaCreateChannelDesc< uchar4 >( );
	cudaMallocArray( &imgArray, &channelDesc1, img.width( ), img.height( ) ); 
	cudaMemcpyToArray( imgArray, 0, 0, &img[ 0 ], img.size( ) * sizeof( uchar4 ), cudaMemcpyHostToDevice );
	cudaBindTextureToArray( imgTex, imgArray, channelDesc1 );

	// GPUɎQƉ摜p̃̈mۂCeNX`Ƀ}bsO
	cudaChannelFormatDesc channelDesc2 = cudaCreateChannelDesc< uchar4 >( );
	cudaMallocArray( &refArray, &channelDesc2, ref.width( ), ref.height( ) ); 
	cudaMemcpyToArray( refArray, 0, 0, &ref[ 0 ], ref.size( ) * sizeof( uchar4 ), cudaMemcpyHostToDevice );
	cudaBindTextureToArray( refTex, refArray, channelDesc2 );

	// `ԂLɂ
	//imgTex.filterMode = cudaFilterModeLinear;
	refTex.filterMode = cudaFilterModeLinear;
	refTex.normalized = 1;

	// CPȔo͉摜IuWFNgp̕ϐ
	mist::array2< float > error( img.width( ), img.height( ) );
	mist::array2< float > scale( img.width( ), img.height( ) );

	// GPȔo͉摜IuWFNgp̕ϐ
	float *pError = NULL, *pScale = NULL;
	cudaMalloc( ( void ** )&pError, img.size( ) * sizeof( float ) );
	cudaMalloc( ( void ** )&pScale, img.size( ) * sizeof( float ) );

	// CPUGPȔo͗p摜
	error.fill( 1.0e10f );
	cudaMemcpy( pError, &error[ 0 ], img.size( ) * sizeof( float ), cudaMemcpyHostToDevice );

	float MSCALE = MAX_SCALE;
    if( img.width( ) * ref.height( ) <  ref.width( ) * img.height( ) )
    {
        MSCALE = static_cast< float >( img.width( ) ) / ref.width( );
    }
    else
    {
        MSCALE = static_cast< float >( img.height( ) ) / ref.height( );
    }

	MSCALE = MAX_SCALE < MSCALE ? MAX_SCALE : MSCALE;

	// ԌvJn
	mist::timer t;

	float s = MIN_SCALE;
	while( s < MSCALE )
	{
		printf( "Scale = %f ...\n", s );

		// sGPŨXbhw
		int W = ( int )img.width( );
		int H = ( int )img.height( );
		int w = ( int )( ref.width( ) * s );
		int h = ( int )( ref.height( ) * s );
		int width  = W - w;
		int height = H - h;
		int threadNumX = 16;
		int threadNumY = 16;
		int blockNumX = width  / threadNumX + ( ( width  % threadNumX ) == 0 ? 0 : 1 );
		int blockNumY = height / threadNumY + ( ( height % threadNumY ) == 0 ? 0 : 1 );
		dim3 nThreads( threadNumX, threadNumY, 1 );
		dim3 nBlocks( blockNumX, blockNumY, 1 );

		// GPUŏs
		kernel<<< nBlocks, nThreads >>>( pError, pScale, W, H, width, height, w, h, s );

		// ܂őҋ@
		cudaThreadSynchronize( );

		// XP[ύX
		s *= SCALE_FACTOR;
	}

	// GPȔʂ]
	cudaMemcpy( &error[ 0 ], pError, error.size( ) * sizeof( float ), cudaMemcpyDeviceToHost );
	cudaMemcpy( &scale[ 0 ], pScale, scale.size( ) * sizeof( float ), cudaMemcpyDeviceToHost );

	float merr = 1.0e10f, mscale = MIN_SCALE;
	int X = 0, Y = 0;
	for( size_t j = 0 ; j < error.height( ) ; j++ )
	{
		for( size_t i = 0 ; i < error.width( ) ; i++ )
		{
			if( merr > error( i, j ) )
			{
				merr = error( i, j );
				mscale = scale( i, j );
				X = i;
				Y = j;
			}
		}
	}

	printf( "Computation time: %.3f\n", t.elapse( ) );

	printf( "\nScale = %f, Err: %.1f at ( %d, %d )\n", mscale, merr, X, Y );

	// ̌n
	cudaUnbindTexture( imgTex );
	cudaUnbindTexture( refTex );
	cudaFreeArray( imgArray );
	cudaFreeArray( refArray );
	cudaFree( pError );
	cudaFree( pScale );

	mist::draw_rect( img, X, Y, X + ( int )( ref.width( ) * mscale ), Y + ( int )( ref.height( ) * mscale ), pixel_type( 255, 0, 0 ) );

	// ʂ̉摜ۑ
	if( !mist::write_bmp( img, argv[ 3 ] ) )
	{
		printf( "Failed to save image to [%s].\n", argv[ 3 ] );
		return( 1 );
	}

	return( 0 );
}
