#ifndef __INCLUDE_CUDA_VECTOR2_OPERATIONS__
#define __INCLUDE_CUDA_VECTOR2_OPERATIONS__

#if !defined(__INCLUDE_CUDA_VECTOR_OPERATIONS__) && !defined(__CUDACC__) && !defined(__CUDABE__)
	#include <math.h>
	#include <vector_types.h>
	#include <vector_functions.h>
#endif

inline __host__ __device__ float outer( float2 a, float2 b )
{
	return( a.x * b.y - a.y * b.x );
}

inline __host__ __device__ float inner( float2 a, float2 b )
{
	return( a.x * b.x + a.y * b.y );
}

inline __host__ __device__ float length( float2 a )
{
	return( sqrtf( inner( a, a ) ) );
}

inline __host__ __device__ float2 operator -( float2 a )
{
	return( make_float2( -a.x, -a.y ) );
}

inline __host__ __device__ float2 operator +( float2 a, float2 b )
{
	return( make_float2( a.x + b.x, a.y + b.y ) );
}

inline __host__ __device__ float2 operator -( float2 a, float2 b )
{
	return( make_float2( a.x - b.x, a.y - b.y ) );
}

inline __host__ __device__ float2 operator *( float2 a, float2 b )
{
	return( make_float2( a.x * b.x, a.y * b.y ) );
}

inline __host__ __device__ float2 operator *( float2 a, float b )
{
	return( make_float2( a.x * b, a.y * b ) );
}

inline __host__ __device__ float2 operator /( float2 a, float b )
{
	return( make_float2( a.x / b, a.y / b ) );
}

inline __host__ __device__ float2 &operator +=( float2 &a, float2 b )
{
	a.x += b.x;
	a.y += b.y;
	return( a );
}

inline __host__ __device__ float2 &operator -=( float2 &a, float2 b )
{
	a.x -= b.x;
	a.y -= b.y;
	return( a );
}

inline __host__ __device__ float2 &operator *=( float2 &a, float2 b )
{
	a.x *= b.x;
	a.y *= b.y;
	return( a );
}

inline __host__ __device__ float2 &operator /=( float2 &a, float2 b )
{
	a.x /= b.x;
	a.y /= b.y;
	return( a );
}

inline __host__ __device__ float2 &operator *=( float2 &a, float b )
{
	a.x *= b;
	a.y *= b;
	return( a );
}

inline __host__ __device__ float2 &operator /=( float2 &a, float b )
{
	a.x /= b;
	a.y /= b;
	return( a );
}

inline __host__ __device__ float2 operator *( float a, float2 b )
{
	return( make_float2( a * b.x, a * b.y ) );
}

#if defined(__CUDACC__) || defined(__CUDABE__)
inline __host__ __device__ float2 unit( float2 a )
{
	return( a * rsqrtf( inner( a, a ) ) );
}
#else
inline __host__ __device__ float2 unit( float2 a )
{
	return( a / sqrtf( inner( a, a ) ) );
}
#endif

#endif	// __INCLUDE_CUDA_VECTOR2_OPERATIONS__
