#ifndef __INCLUDE_CUDA_VECTOR3_OPERATIONS__
#define __INCLUDE_CUDA_VECTOR3_OPERATIONS__

#if !defined(__INCLUDE_CUDA_VECTOR_OPERATIONS__) && !defined(__CUDACC__) && !defined(__CUDABE__)
	#include <math.h>
	#include <vector_types.h>
	#include <vector_functions.h>
#endif

inline __host__ __device__ float3 outer( float3 a, float3 b )
{
	return( make_float3( a.y * b.z - a.z * b.y, a.z * b.x - a.x * b.z, a.x * b.y - a.y * b.x ) );
}

inline __host__ __device__ float inner( float3 a, float3 b )
{
	return( a.x * b.x + a.y * b.y + a.z * b.z );
}

inline __host__ __device__ float length( float3 a )
{
	return( sqrtf( inner( a, a ) ) );
}

inline __host__ __device__ float3 operator -( float3 a )
{
	return( make_float3( -a.x, -a.y, -a.z ) );
}

inline __host__ __device__ float3 operator +( float3 a, float3 b )
{
	return( make_float3( a.x + b.x, a.y + b.y, a.z + b.z ) );
}

inline __host__ __device__ float3 operator -( float3 a, float3 b )
{
	return( make_float3( a.x - b.x, a.y - b.y, a.z - b.z ) );
}

inline __host__ __device__ float3 operator *( float3 a, float3 b )
{
	return( make_float3( a.x * b.x, a.y * b.y, a.z * b.z ) );
}

inline __host__ __device__ float3 operator *( float3 a, float b )
{
	return( make_float3( a.x * b, a.y * b, a.z * b ) );
}

inline __host__ __device__ float3 operator /( float3 a, float b )
{
	return( make_float3( a.x / b, a.y / b, a.z / b ) );
}

inline __host__ __device__ float3 &operator +=( float3 &a, float3 b )
{
	a.x += b.x;
	a.y += b.y;
	a.z += b.z;
	return( a );
}

inline __host__ __device__ float3 &operator -=( float3 &a, float3 b )
{
	a.x -= b.x;
	a.y -= b.y;
	a.z -= b.z;
	return( a );
}

inline __host__ __device__ float3 &operator *=( float3 &a, float3 b )
{
	a.x *= b.x;
	a.y *= b.y;
	a.z *= b.z;
	return( a );
}

inline __host__ __device__ float3 &operator /=( float3 &a, float3 b )
{
	a.x /= b.x;
	a.y /= b.y;
	a.z /= b.z;
	return( a );
}

inline __host__ __device__ float3 &operator *=( float3 &a, float b )
{
	a.x *= b;
	a.y *= b;
	a.z *= b;
	return( a );
}

inline __host__ __device__ float3 &operator /=( float3 &a, float b )
{
	a.x /= b;
	a.y /= b;
	a.z /= b;
	return( a );
}

inline __host__ __device__ float3 operator *( float a, float3 b )
{
	return( make_float3( a * b.x, a * b.y, a * b.z ) );
}

#if defined(__CUDACC__) || defined(__CUDABE__)
inline __host__ __device__ float3 unit( float3 a )
{
	return( a * rsqrtf( inner( a, a ) ) );
}
#else
inline __host__ __device__ float3 unit( float3 a )
{
	return( a / sqrtf( inner( a, a ) ) );
}
#endif

#endif	// __INCLUDE_CUDA_VECTOR3_OPERATIONS__
