#ifndef __INCLUDE_CUDA_VECTOR4_OPERATIONS__
#define __INCLUDE_CUDA_VECTOR4_OPERATIONS__

#if !defined(__INCLUDE_CUDA_VECTOR_OPERATIONS__) && !defined(__CUDACC__) && !defined(__CUDABE__)
	#include <math.h>
	#include <vector_types.h>
	#include <vector_functions.h>
#endif

//inline __host__ __device__ float4 outer( float4 a, float4 b )
//{
//	return( make_float4( a.y * b.z - a.z * b.y, a.z * b.x - a.x * b.z, a.x * b.y - a.y * b.x ) );
//}

inline __host__ __device__ float inner( float4 a, float4 b )
{
	return( a.x * b.x + a.y * b.y + a.z * b.z + a.w * b.w );
}

inline __host__ __device__ float length( float4 a )
{
	return( sqrtf( inner( a, a ) ) );
}

inline __host__ __device__ float4 operator -( float4 a )
{
	return( make_float4( -a.x, -a.y, -a.z, -a.w ) );
}

inline __host__ __device__ float4 operator +( float4 a, float4 b )
{
	return( make_float4( a.x + b.x, a.y + b.y, a.z + b.z, a.w + b.w ) );
}

inline __host__ __device__ float4 operator -( float4 a, float4 b )
{
	return( make_float4( a.x - b.x, a.y - b.y, a.z - b.z, a.w - b.w ) );
}

inline __host__ __device__ float4 operator *( float4 a, float4 b )
{
	return( make_float4( a.x * b.x, a.y * b.y, a.z * b.z, a.w * b.w ) );
}

inline __host__ __device__ float4 operator *( float4 a, float b )
{
	return( make_float4( a.x * b, a.y * b, a.z * b, a.w * b ) );
}

inline __host__ __device__ float4 operator /( float4 a, float b )
{
	return( make_float4( a.x / b, a.y / b, a.z / b, a.w / b ) );
}

inline __host__ __device__ float4 &operator +=( float4 &a, float4 b )
{
	a.x += b.x;
	a.y += b.y;
	a.z += b.z;
	a.w += b.w;
	return( a );
}

inline __host__ __device__ float4 &operator -=( float4 &a, float4 b )
{
	a.x -= b.x;
	a.y -= b.y;
	a.z -= b.z;
	a.w -= b.w;
	return( a );
}

inline __host__ __device__ float4 &operator *=( float4 &a, float4 b )
{
	a.x *= b.x;
	a.y *= b.y;
	a.z *= b.z;
	a.w *= b.w;
	return( a );
}

inline __host__ __device__ float4 &operator /=( float4 &a, float4 b )
{
	a.x /= b.x;
	a.y /= b.y;
	a.z /= b.z;
	a.w /= b.w;
	return( a );
}

inline __host__ __device__ float4 &operator *=( float4 &a, float b )
{
	a.x *= b;
	a.y *= b;
	a.z *= b;
	a.w *= b;
	return( a );
}

inline __host__ __device__ float4 &operator /=( float4 &a, float b )
{
	a.x /= b;
	a.y /= b;
	a.z /= b;
	a.w /= b;
	return( a );
}

inline __host__ __device__ float4 operator *( float a, float4 b )
{
	return( make_float4( a * b.x, a * b.y, a * b.z, a * b.w ) );
}

#if defined(__CUDACC__) || defined(__CUDABE__)
inline __host__ __device__ float4 unit( float4 a )
{
	return( a * rsqrtf( inner( a, a ) ) );
}
#else
inline __host__ __device__ float4 unit( float4 a )
{
	return( a / sqrtf( inner( a, a ) ) );
}
#endif

#endif	// __INCLUDE_CUDA_VECTOR4_OPERATIONS__
